import java.io.*;
import java.util.*;

public class ImportDatMap { 
    static final int SPEED[] = {60, 40, 20}; // average speed km/h for streets at different levels

    static String[] split(String line, int n) {
        String[] columns = new String[n];
        int q = 0;
        for (int i = 0; i < n; i++) { 
            int p = line.indexOf(',', q);
            columns[i] = p < 0 ? line.substring(q) : line.substring(q, p);
            q = p + 1;
        }
        return columns;
    }

    public static void main(String[] args) throws Exception { 
        BufferedReader in;
        String line;
        TreeMap streets = new TreeMap();
        String dir = args[0];

        // read street data
        in = new BufferedReader(new FileReader(new File(dir, "street.dat")));
        Street street = null;
        while ((line = in.readLine()) != null) {
            String[] data = split(line, 6);
            String label = data[0];//data[1];
            int streetLevel = Integer.parseInt(data[2])-1;
            Segment s;
            double x = Double.parseDouble(data[4]);
            double y = Double.parseDouble(data[5]);
            if (street == null || !street.name.equals(label)) { 
                street = (Street)streets.get(label);
            }
            if (street == null) { 
                street = new Street(label, streetLevel, ObjectType.STREET);
                streets.put(label, street);  
                s = new Segment(street);
                s.oneWay = Integer.parseInt(data[3]) == 1;
                street.segments.add(s);
            } else { 
                s = (Segment)street.segments.get(street.segments.size()-1);
                Point lastPoint = (Point)s.points.get(s.points.size()-1);
                if (x == lastPoint.x && y == lastPoint.y) { 
                    continue;
                }
                if (s.points.size() == 0xFF) { 
                    Segment newSegment = new Segment(street);
                    newSegment.points.add(s.points.get(0xFE));
                    newSegment.oneWay = s.oneWay;
                    s = newSegment;
                    street.segments.add(s);
                }
            }
            s.points.add(new Point(x, y, s));
            /*
            int p, q = line.indexOf(',');
            String label = line.substring(1, q-1);
            street = new Street(label, streetLevel, ObjectType.STREET);
            streets.put(label, street);  
            s = new Segment(street);
            street.segments.add(s);
            levels[streetLevel].add(s);
double px = 0, py = 0;
lineNo += 1;
            while ((p = line.indexOf('(', q)) >= 0) { 
                q = line.indexOf(',', p);
                double x = Double.parseDouble(line.substring(p+1, q));
                p = line.indexOf(')', q);                
                double y = Double.parseDouble(line.substring(q+1, p));
                q = p + 1;
                if (x != px || y != py) { 
                    if (s.points.size() == 0xFF) { 
                        Segment newSegment = new Segment(street);
                        newSegment.points.add(s.points.get(0xFE));
                        s = newSegment;
                        street.segments.add(s);
                        levels[streetLevel].add(s);
                    }
                    px = x;
                    py = y;
                    s.points.add(new Point(x, y, s));
                }
            }
            */
        }
        in.close();

        for (int i = ObjectType.FIRST_OBJECT; i <= ObjectType.LAST_OBJECT; i++) { 
            File f = new File(dir, ObjectType.name[i] + ".dat");
            if (f.exists()) { 
                in = new BufferedReader(new FileReader(f));
                while ((line = in.readLine()) != null) {
                    String[] data = split(line, 4);        
                    int streetLevel = 0;
                    String label = data[2] + '\n' + data[3];
                    street = new Street(label, streetLevel, i);
                    streets.put(label, street);  
                    Segment s = new Segment(street);
                    street.segments.add(s);
                    double x = Double.parseDouble(data[0]);
                    double y = Double.parseDouble(data[1]);
                    s.points.add(new Point(x, y, s));
                }
                in.close();
            }
        }

        for (int i = ObjectType.FIRST_REGION; i <= ObjectType.LAST_REGION; i++) { 
            File f = new File(dir, ObjectType.name[i] + ".dat");
            if (f.exists()) { 
                in = new BufferedReader(new FileReader(f));
                while ((line = in.readLine()) != null) {
                    String[] data = split(line, 3);        
                    String label = ObjectType.name[i] + data[0];
                    street = (Street)streets.get(label);
                    Segment s;
                    int streetLevel = 0;
                    if (street == null) { 
                        street = new Street(label, streetLevel, i);
                        streets.put(label, street);  
                        s = new Segment(street);
                        street.segments.add(s);
                    } else { 
                        s = (Segment)street.segments.get(street.segments.size()-1);
                        if (s.points.size() == 0xFE) { 
                            s.points.add(s.points.get(0));
                            Segment newSegment = new Segment(street);
                            newSegment.points.add(s.points.get(0));
                            newSegment.points.add(s.points.get(0xFD));
                            s = newSegment;
                            street.segments.add(s);
                        }
                    }                        
                    double x = Double.parseDouble(data[1]);
                    double y = Double.parseDouble(data[2]);
                    s.points.add(new Point(x, y, s));
                }
                in.close();
            }
        }
        MapWriter.write(streets.size(), streets, new MetricGrid(), SPEED, true, false);
    }
}