import java.io.*;
import java.util.*;

public class MapWriter { 
    static final String OUTPUT_MAP_FILE = "..\\res\\city.map";
    static final int CELL_BITS = 3;
    static final int N_CELLS = 1 << CELL_BITS;
    static final int MAX_COORD_BITS = 16;
    static final int MAX_COORD = (1 << MAX_COORD_BITS) - 1;

    static final double MIN_DISTANCE = 1;

    static void mapPoint(Point p, double minX, double minY, double scaleX, double scaleY) {
        p.sx = (int)((p.x - minX) / scaleX * MAX_COORD);
        p.sy = ~(int)((p.y - minY) / scaleY * MAX_COORD) & 0xFFFF;
    }

    static void addCrossPoints(Collection streets, Grid grid, double minX, double minY, double scaleX, double scaleY) { 
        Iterator i1 = streets.iterator();
        while (i1.hasNext()) { 
            for (Street s1 = (Street)i1.next(); s1 != null; s1 = s1.collision) {
                if (s1.type != ObjectType.STREET) { 
                    continue;
                }
                for (int j1 = 0, ns1 = s1.segments.size(); j1 < ns1; j1++) { 
                    Segment seg1 = (Segment)s1.segments.get(j1);
                    Point sp1 = (Point)seg1.points.get(0), fp1;
                    double sx1 = sp1.x, fx1;
                    double sy1 = sp1.y, fy1;
                    for (int k1 = 1; k1 < seg1.points.size(); k1++, sx1 = fx1, sy1 = fy1, sp1 = fp1) { 
                        fp1 = (Point)seg1.points.get(k1);
                        fx1 = fp1.x;
                        fy1 = fp1.y;

                        Iterator i2 = streets.iterator();
                        while (i2.hasNext()) { 
                            for (Street s2 = (Street)i2.next(); s2 != null; s2 = s2.collision) { 
                                if (s1 != s2 && s2.type == ObjectType.STREET) { 
                                    for (int j2 = 0, ns2 = s2.segments.size(); j2 < ns2; j2++) { 
                                        Segment seg2 = (Segment)s2.segments.get(j2);
                                        Point sp2 = (Point)seg2.points.get(0), fp2;
                                        double sx2 = sp2.x, fx2;
                                        double sy2 = sp2.y, fy2;

                                        for (int k2 = 1; k2 < seg2.points.size(); k2++, sx2 = fx2, sy2 = fy2, sp2 = fp2) {
                                            fp2 = (Point)seg2.points.get(k2);
                                            fx2 = fp2.x;
                                            fy2 = fp2.y;

                                            if (sp1.equals(sp2) || sp1.equals(fp2) || fp1.equals(sp2) || fp1.equals(fp2))
                                            {
                                                continue;
                                            }

                                            double x, y;

                                            if (sy1 == fy1 || sy2 == fy2) { 
                                                if (sy1 != fy1) { 
                                                    y = sy2;
                                                    x = (y-sy1)*(fx1-sx1)/(fy1-sy1) + sx1;
                                                } else if (sy2 != fy2) { 
                                                    y = sy1;
                                                    x = (y-sy2)*(fx2-sx2)/(fy2-sy2) + sx2;
                                                } else { 
                                                    continue;
                                                }
                                            } else { 
                                                double t = (fx1-sx1)*(fy2-sy2) - (fx2-sx2)*(fy1-sy1);
                                                if (t == 0) { 
                                                    continue;
                                                }
                                                y = ((sx2-sx1)*(fy1-sy1)*(fy2-sy2) - sy2*(fx2-sx2)*(fy1-sy1) + sy1*(fx1-sx1)*(fy2-sy2)) / t;
                                                x = (fx1-sx1)*(y-sy1)/(fy1-sy1) + sx1;
                                            }

                                            Point p = new Point(x, y, seg1);
                                            mapPoint(p, minX, minY, scaleX, scaleY);
                                            boolean inx1 = (p.sx >= sp1.sx && p.sx <= fp1.sx) || (p.sx >= fp1.sx && p.sx <= sp1.sx);
                                            boolean inx2 = (p.sx >= sp2.sx && p.sx <= fp2.sx) || (p.sx >= fp2.sx && p.sx <= sp2.sx);
                                            boolean iny1 = (p.sy >= sp1.sy && p.sy <= fp1.sy) || (p.sy >= fp1.sy && p.sy <= sp1.sy);
                                            boolean iny2 = (p.sy >= sp2.sy && p.sy <= fp2.sy) || (p.sy >= fp2.sy && p.sy <= sp2.sy);
                                            if (inx1 && inx2 && iny1 && iny2) { 
                                                if (!p.equals(sp1) && !p.equals(fp1)) { 
                                                    //System.out.println("Add cross point1 (" + x + "," + y + ") for intervals (" + sx1 + "," + sy1 + ")-(" + fx1 + "," + fy1 + ") and (" + sx2 + "," + sy2 + ")-(" + fx2 + "," + fy2 + ")");
                                                    p.segment = seg1;
                                                    seg1.points.add(k1, p);
                                                    fx1 = x;
                                                    fy1 = y;
                                                    fp1 = p;
                                                }
                                                if (!p.equals(sp2) && !p.equals(fp2)) { 
                                                    //System.out.println("Add cross point2 (" + x + "," + y + ") for intervals (" + sx1 + "," + sy1 + ")-(" + fx1 + "," + fy1 + ") and (" + sx2 + "," + sy2 + ")-(" + fx2 + "," + fy2 + ")");
                                                    p.segment = seg2;
                                                    seg2.points.add(k2, p);
                                                    fx2 = x;
                                                    fy2 = y;
                                                    fp2 = p;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
                                                

    static void write(int nStreets, TreeMap streets, Grid grid, 
                      int[] speed, boolean crosses, boolean small) 
        throws Exception
    {
        Iterator iterator;
        ArrayList levelList = new ArrayList();
        iterator = streets.values().iterator();
        for (int i = 0; iterator.hasNext();) { 
            Street chain = (Street)iterator.next();
            for (Street s = chain; s != null; s = s.collision) {                 
                while (s.level >= levelList.size()) { 
                    levelList.add(new ArrayList());
                }
                ArrayList level = (ArrayList)levelList.get(s.level);
                s.index = i++;
                for (int j = 0, ns = s.segments.size(); j < ns; j++) {                     
                    if (s.type >= ObjectType.FIRST_REGION && s.type <= ObjectType.LAST_REGION) { 
                        level.add(0, s.segments.get(j));
                    } else { 
                        level.add(s.segments.get(j));
                    }
                }
            }
        }
        ArrayList[] levels = (ArrayList[])levelList.toArray(new ArrayList[levelList.size()]);
        int nLevels = levels.length;
                                                            
        double minX = Double.MAX_VALUE;
        double maxX = Double.MIN_VALUE;
        double minY = Double.MAX_VALUE;
        double maxY = Double.MIN_VALUE;
        int nPoints = 0;
        int totalNameLength = 0;


        DataOutputStream out = new DataOutputStream(new FileOutputStream(OUTPUT_MAP_FILE));
        out.writeShort(nStreets);
        iterator = streets.values().iterator();
        while (iterator.hasNext()) { 
            Street chain = (Street)iterator.next();
            for (Street s = chain; s != null; s = s.collision) {                 
                for (int j = 0, ns = s.segments.size(); j < ns; j++) { 
                    Segment seg = (Segment)s.segments.get(j);
                    for (int k = 0, np = seg.points.size(); k < np; k++) { 
                        Point p = (Point)seg.points.get(k);
                        double x = p.x;
                        double y = p.y;
                        if (x < minX) { 
                            minX = x;
                        } 
                        if (x > maxX) { 
                            maxX = x;
                        } 
                        if (y < minY) { 
                            minY = y;
                        } 
                        if (y > maxY) { 
                            maxY = y;
                        }               
                    }
                }                    
                out.writeByte(s.segments.size());
                out.writeByte(s.type);
                byte[] nameBytes = s.name.getBytes();
                out.writeByte(nameBytes.length);
                totalNameLength += nameBytes.length + 1;
                out.write(nameBytes);
           }
        }

        double scaleX = maxX - minX;
        double scaleY = maxY - minY;

        for (int i = 0; i < nLevels; i++) { 
            ArrayList list = levels[i];
            for (int j = 0, n = list.size(); j < n; j++) {
                Segment s = (Segment)list.get(j);
                for (int k = 0, np = s.points.size(); k < np; k++) {                 
                    Point p = (Point)s.points.get(k);
                    mapPoint(p, minX, minY, scaleX, scaleY);
                }
            }
        }
        if (small || crosses) { 
            addCrossPoints(streets.values(), grid, minX, minY, scaleX, scaleY);
        }

        System.out.println("MinX = " + minX);
        System.out.println("MaxX = " + maxX);
        System.out.println("MinY = " + minY);
        System.out.println("MaxY = " + maxY);

        int cells[][] = new int[N_CELLS][N_CELLS];
        HashMap crossMap = new HashMap();
        int nCrosses = 0;
    
        for (int i = 0; i < nLevels; i++) { 
            ArrayList list = levels[i];
            for (int j = 0, n = list.size(); j < n; j++) {
                Segment s = (Segment)list.get(j);
                for (int k = 0, np = s.points.size(); k < np; k++) {                 
                    Point p = (Point)s.points.get(k);
                    if ((small || crosses) && s.street.type == ObjectType.STREET) {
                        Cross c = (Cross)crossMap.get(p);
                        if (c == null) { 
                            c = new Cross(p.sx, p.sy);
                            crossMap.put(p, c);
                            if (k == 0 || k+1 == np) { 
                                nCrosses += 1;
                                c.intermediate = false;
                            }
                        } else if (c.intermediate) { 
                            nCrosses += 1;
                            c.intermediate = false;
                        }
                        p.cross = c;
                        c.segments.add(s);
                    }
                }
            }
        }

        if (small) {
            for (int i = 0; i < nLevels; i++) { 
                ArrayList list = levels[i];
                for (int j = 0, n = list.size(); j < n; j++) {            
                    Segment s = (Segment)list.get(j);
                    if (s.street.type == ObjectType.STREET) {
                        int k = 0; 
                        while (k < s.points.size()) {                 
                            Point p = (Point)s.points.get(k); 
                            if (k != 0 && k+1 != s.points.size() && ((Cross)crossMap.get(p)).intermediate) {
                                s.points.remove(k);
                            } else { 
                                k += 1;
                            }
                        }
                    }
                }
            }
        }


        for (int i = 0; i < nLevels; i++) { 
            ArrayList list = levels[i];
            for (int j = 0, n = list.size(); j < n; j++) {
                Segment s = (Segment)list.get(j);
                int minSX = Integer.MAX_VALUE;
                int maxSX = Integer.MIN_VALUE;
                int minSY = Integer.MAX_VALUE;
                int maxSY = Integer.MIN_VALUE;
                for (int k = 0, np = s.points.size(); k < np; k++) {                 
                    Point p = (Point)s.points.get(k);
                    nPoints += 1;
                    int x = p.sx;
                    int y = p.sy;
                    if (x < minSX) { 
                        minSX = x;
                    } 
                    if (x > maxSX) { 
                        maxSX = x;
                    } 
                    if (y < minSY) { 
                        minSY = y;
                    } 
                    if (y > maxSY) { 
                        maxSY = y;
                    }
                }
                minSX >>>= MAX_COORD_BITS-CELL_BITS;
                maxSX >>>= MAX_COORD_BITS-CELL_BITS;
                minSY >>>= MAX_COORD_BITS-CELL_BITS;
                maxSY >>>= MAX_COORD_BITS-CELL_BITS;
                for (int cx = minSX; cx <= maxSX; cx++) { 
                    for (int cy = minSY; cy <= maxSY; cy++) { 
                        cells[cx][cy] += 1;
                    }
                }
            }
        }
        for (int i = 0; i < N_CELLS; i++) { 
            for (int j = 0; j < N_CELLS; j++) { 
                out.writeShort(cells[i][j]);
            }
        }
        out.writeByte(nLevels);            

        for (int i = 0; i < nLevels; i++) { 
            ArrayList list = levels[i];
            out.writeShort(list.size());      
            for (int j = 0, n = list.size(); j < n; j++) {            
                Segment s = (Segment)list.get(j);
                out.writeShort(s.street.index);
                out.writeByte(s.points.size());
                for (int k = 0, np = s.points.size(); k < np; k++) {                 
                    Point p = (Point)s.points.get(k); 
                    out.writeShort(p.sx);
                    out.writeShort(p.sy);
                }
            }
        }

        if (crosses) {
            out.writeShort(nCrosses);
            System.out.println("Write cross data " + nCrosses);
            iterator = crossMap.values().iterator();
            Cross[] ca = new Cross[nCrosses];
            for (int i = 0; iterator.hasNext();) { 
                Cross c = (Cross)iterator.next();
                if (!c.intermediate) {
                    ca[i] = c;
                    c.index = i++;
                    ArrayList neighbors = new ArrayList();
                    for (int j = 0; j < c.segments.size(); j++) { 
                        Segment s = (Segment)c.segments.get(j);
                        int k = 0; 
                        while (((Point)s.points.get(k)).cross != c) { 
                            k += 1;
                        }
                        Point pk = (Point)s.points.get(k);
                        Point p0 = pk;
                        double distance = 0;
                        double time = 0;
                        if (!s.oneWay) { 
                            for (int n = k; --n >= 0;) {
                                Point p = (Point)s.points.get(n);
                                double d = grid.distance(p, p0); 
                                p0 = p;
                                distance += d;
                                time += d*3.6/speed[s.street.level]; 
                                if (!p.cross.intermediate) { 
                                    neighbors.add(new Edge(s, p.cross, (int)distance, (int)time));
                                    break;
                                } 
                            }
                            distance = 0; 
                            time = 0;
                            p0 = pk;
                        } 
                        for (int n = k; ++n < s.points.size();) {
                            Point p = (Point)s.points.get(n);
                            double d = grid.distance(p, p0);
                            p0 = p;
                            distance += d;
                            time += d*3.6/speed[s.street.level]; 
                            if (!p.cross.intermediate) { 
                                neighbors.add(new Edge(s, p.cross, (int)distance, (int)time));
                                break;
                            }
                        }
                    }
                    c.neighbors = (Edge[])neighbors.toArray(new Edge[neighbors.size()]);
                }
            }
            for (int i = 0; i < ca.length; i++) { 
                Cross c = ca[i];
                out.writeShort(c.x);
                out.writeShort(c.y);                    
                out.writeByte(c.neighbors.length);
                for (int j = 0; j < c.neighbors.length; j++) { 
                    Edge e = c.neighbors[j];
                    out.writeShort(e.target.index);
                    out.writeShort(e.segment.street.index);
                    out.writeByte(e.segment.street.segments.indexOf(e.segment));
                    out.writeShort(e.distance);
                    out.writeShort(e.time);
                }
            }
        } else { 
            out.writeShort(0);
        }                        
        out.close();

        System.out.println("Number of levels: " + nLevels + 
                           "\nNumber of streets: " + nStreets + 
                           "\nNumber of points: " + nPoints + 
                           "\nTotal name length: " + totalNameLength);
    }
}