import javax.microedition.lcdui.*;
import java.util.*;

public class MapCanvas extends Canvas implements CommandListener {
    MobileMap  main;
    int        screenWidth;
    int        screenHeight;
    int        x, y, size;
    int        cx, cy;
    int        scale;
    int        initScale;
    int        level; 
    short      marker;
    int        selectedStreetIndex;
    int        printNamesThreshold;
    int[]      minLevelScale;
    Font       font;
    int        fontHeight;
    int        fontSpaceWidth;
    Timer      repeatTimer;
    RepeatTask repeatTask;
    Route      route;
    Image      image;
    int        indent;
    int        imageX, imageY, imageScale;    

    byte[]     matrix;
    int        matrixColumns;
    int        matrixRows;

    int        fireMode;
    static final int FIRE_INFO         = 0;
    static final int FIRE_START_POINT  = 1;
    static final int FIRE_FINISH_POINT = 2;
    
    static final int FONT_HEIGHT_INTERLACING = 70;

    Cross      startPoint;
    Cross      finishPoint;


    static final int MATRIX_CELL_BITS = 3;
    static final int MAX_COORD_BITS = 16;
    static final int MAX_COORD = (1 << MAX_COORD_BITS) - 1;
    static final int STREETS_PER_SCREEN = 20;
    static final int ICONS_PER_SCREEN = 40;
    static final int MAX_PIECES = 64;

    static final int BACKGROUND_COLOR = 0xf0bf73; //0xff9f34;
    static final int CURSOR_COLOR     = 0x00ffff; 
    static final int SELECTION_COLOR  = 0xff00ff;
    static final int ROUTE_COLOR      = 0x00ff80;
    static final int OBJECT_COLOR     = 0xff0000;
    static final int NAME_COLOR       = 0x000000;
    static final int ROAD_COLOR       = 0xffffff;
    static final int GREEN_COLOR      = 0x00ff00;
    static final int WATER_COLOR      = 0x0000ff;
    static final int BUILDING_COLOR   = 0x804000;

    static final int REPEAT_DELAY = 1000;
    static final int REPEAT_PERIOD = 100;

    static final int CURSOR_STEP = 4;
    static final int CURSOR_SIZE = 4;

    static final int FLAG_SIZE = 10;

    MapCanvas(MobileMap main) { 
        setFullScreenMode(true);
        this.main = main;
        repeatTimer = new Timer();
 	screenWidth = getWidth();
	screenHeight = getHeight();
        initScale = screenWidth < screenHeight ? screenWidth : screenHeight; 
        minLevelScale = new int[main.nLevels];
        int nStreets = main.nLevelStreets[0];
        int scale = 1;
        for (int i = 1; i < minLevelScale.length; i++) { 
            nStreets += main.nLevelStreets[i];
            while (scale*scale < nStreets / STREETS_PER_SCREEN) {
                scale += 1;
            }
            minLevelScale[i] = initScale * scale;
        }
        if (main.nLevels == 1) { 
            while (scale*scale < nStreets / STREETS_PER_SCREEN) {
                scale += 1;
            }            
            printNamesThreshold = initScale * scale / 2; 
            // if there is only one level, start printing names earlier
        } else { 
            printNamesThreshold = minLevelScale[1];
        }
        font = Font.getFont(Font.FACE_PROPORTIONAL/*SYSTEM*/, Font.STYLE_PLAIN, Font.SIZE_SMALL);
        fontHeight = font.getHeight()*FONT_HEIGHT_INTERLACING/100;
        indent = 0;
        fontSpaceWidth = font.charWidth(' ');
        reset();
        setCommandListener(this);
        addCommand(MobileMap.SEARCH_CMD);
        addCommand(MobileMap.INDEX_CMD);
        addCommand(MobileMap.ZOOM_IN_CMD);
        addCommand(MobileMap.ZOOM_OUT_CMD);

        if (main.crosses.length != 0) { 
            addCommand(MobileMap.START_POINT_CMD);
            addCommand(MobileMap.FINISH_POINT_CMD);
            addCommand(MobileMap.FASTEST_ROUTE_CMD);
            addCommand(MobileMap.SHORTEST_ROUTE_CMD);
        }
        addCommand(MobileMap.RESET_CMD);
        addCommand(MobileMap.HELP_CMD);
        addCommand(MobileMap.QUIT_CMD);
        Display.getDisplay(main).setCurrent(this);
    }

    class RepeatTask extends TimerTask { 
        RepeatTask(int keyCode) { 
            this.keyCode = keyCode;
        }

        public void run() { 
            keyAction(keyCode);
        }

        int keyCode;
    }
    

    protected void paint(Graphics g) {
 	screenWidth = getWidth();
	screenHeight = getHeight();
        if (image == null || imageX != x || imageY != y || imageScale != scale) { 
            if (image == null) { 
                image = Image.createImage(screenWidth, screenHeight);
            }
            if (matrix == null) {
                matrixColumns = ((screenWidth + (1 << (MATRIX_CELL_BITS+3)) - 1) >> (MATRIX_CELL_BITS+3)) + 1;
                matrixRows = ((screenHeight + (1 << MATRIX_CELL_BITS) - 1) >> MATRIX_CELL_BITS) + 1;
                matrix = new byte [matrixColumns*matrixRows];
            } else { 
                for (int i = 0; i < matrix.length; i++) { 
                    matrix[i] = 0;
                }
            }
            imageX = x;
            imageY = y;
            imageScale = scale;
            Graphics bg = image.getGraphics();
            bg.setColor(BACKGROUND_COLOR); 
            bg.fillRect(bg.getClipX(), bg.getClipY(), bg.getClipWidth(), bg.getClipHeight());
            bg.setFont(font);
            marker += 1;
            int x0 = x >>> (MAX_COORD_BITS - MobileMap.CELL_BITS);
            int x1 = Math.min(MAX_COORD, x + (screenWidth << MAX_COORD_BITS) / scale) >>> (MAX_COORD_BITS - MobileMap.CELL_BITS);
            int y0 = y >>> (MAX_COORD_BITS - MobileMap.CELL_BITS);
            int y1 = Math.min(MAX_COORD, y + (screenHeight << MAX_COORD_BITS) / scale) >>> (MAX_COORD_BITS - MobileMap.CELL_BITS);
            for (int i = x0;  i <= x1; i++) {
                for (int j = y0;  j <= y1; j++) {
                    drawCell(i, j, bg, false);
                }
            }
            if (scale >= printNamesThreshold) { 
                marker += 1;
                for (int i = x0; i <= x1; i++) {
                    for (int j = y0; j <= y1; j++) {
                        drawCell(i, j, bg, true);
                    }
                }
            }            
        }
        g.drawImage(image, 0, 0, Graphics.TOP|Graphics.LEFT);
            
        g.setColor(CURSOR_COLOR); 
        g.drawLine(cx-CURSOR_SIZE, cy, cx+CURSOR_SIZE, cy);
        g.drawLine(cx, cy-CURSOR_SIZE, cx, cy+CURSOR_SIZE);   
        if (fireMode == FIRE_START_POINT || fireMode == FIRE_FINISH_POINT) { 
            g.drawArc(cx-CURSOR_SIZE, cy-CURSOR_SIZE, CURSOR_SIZE*2, CURSOR_SIZE*2, 0, 360); 
        }
        if (startPoint != null) { 
            int xp = startPoint.x & MAX_COORD;
            int yp = startPoint.y & MAX_COORD;
            
            if (xp >= x && xp < x + (screenWidth << MAX_COORD_BITS) / scale 
                && yp >= y && yp < y + (screenHeight << MAX_COORD_BITS) / scale) 
            { 
                xp = (xp-x)*scale >> MAX_COORD_BITS;
                yp = (yp-y)*scale >> MAX_COORD_BITS;
                g.setColor(ROUTE_COLOR);                         
                g.drawLine(xp, yp, xp, yp-FLAG_SIZE); 
                g.drawLine(xp, yp-FLAG_SIZE, xp-FLAG_SIZE/2, yp-FLAG_SIZE*3/4); 
                g.drawLine(xp-FLAG_SIZE/2, yp-FLAG_SIZE*3/4, xp, yp-FLAG_SIZE/2); 
            }            
        }
        if (finishPoint != null) { 
            int xp = finishPoint.x & MAX_COORD;
            int yp = finishPoint.y & MAX_COORD;
            
            if (xp >= x && xp < x + (screenWidth << MAX_COORD_BITS) / scale
                && yp >= y && yp < y + (screenHeight << MAX_COORD_BITS) / scale) 
            { 
                xp = (xp-x)*scale >> MAX_COORD_BITS;
                yp = (yp-y)*scale >> MAX_COORD_BITS;
                g.setColor(ROUTE_COLOR);                         
                g.drawLine(xp, yp, xp, yp-FLAG_SIZE); 
                g.drawLine(xp, yp-FLAG_SIZE, xp+FLAG_SIZE/2, yp-FLAG_SIZE*3/4); 
                g.drawLine(xp+FLAG_SIZE/2, yp-FLAG_SIZE*3/4, xp, yp-FLAG_SIZE/2); 
            }
        }            
    }

    int regionColor(Street street) { 
        switch (street.type) { 
        case ObjectType.GREEN:
            return GREEN_COLOR;
        case ObjectType.WATER:
            return WATER_COLOR;
        case ObjectType.BUILDING:
            return BUILDING_COLOR;
        default:
            return BACKGROUND_COLOR;
        }
    }


    final boolean isOccupied(int x, int y) { 
        int mx0 = (x > fontSpaceWidth/2 ? x - fontSpaceWidth/2 : 0) >> MATRIX_CELL_BITS;
        int mx1 = (x + fontSpaceWidth/2) >> MATRIX_CELL_BITS;
        int my0 = (y > fontHeight/2 ? y - fontHeight/2 : 0) >> MATRIX_CELL_BITS;
        int my1 = (y + fontHeight/2) >> MATRIX_CELL_BITS;
        for (int my = my0; my <= my1; my++) { 
            for (int mx = mx0; mx <= mx1; mx++) { 
                if ((matrix[my*matrixColumns + (mx>>3)] & (1 << (mx & 7))) != 0) { 
                    return true;
                }
            }
        }
        return false;
    }

    final void setOccupied(int x, int y) { 
        int mx0 = (x > fontSpaceWidth/2 ? x - fontSpaceWidth/2 : 0) >> MATRIX_CELL_BITS;
        int mx1 = (x + fontSpaceWidth/2) >> MATRIX_CELL_BITS;
        int my0 = (y > fontHeight/2 ? y - fontHeight/2 : 0) >> MATRIX_CELL_BITS;
        int my1 = (y + fontHeight/2) >> MATRIX_CELL_BITS;
        for (int my = my0; my <= my1; my++) { 
            for (int mx = mx0; mx <= mx1; mx++) { 
                matrix[my*matrixColumns + (mx>>3)] |= (1 << (mx & 7));
            }
        }
    }



    void drawCell(int i, int j, Graphics g, boolean printNames) {
        Segment[] segments = main.cells[i][j];
      next_segment:
        for (int k = 0; k < segments.length; k++) { 
            Segment s = segments[k];
            if (s.level > level) { 
                break;
            } 
            if (s.marker != marker) { 
                s.marker = marker;
                Street street = main.streets[s.streetIndex];
                String name = street.name;
                int    nameLen = name.length();
                int    firstPoint = 0;
                if (printNames) { 
                    if (s.maxLength > (font.stringWidth(name) >> 1)) { 
                        for (int l = 0; l < street.segments.length; l++) { 
                            if (street.segments[l].marker == marker-1
                                && street.segments[l].maxLength > s.maxLength) 
                            { 
                                continue next_segment;
                            }
                            street.segments[l].marker = marker;
                        }
                        firstPoint = s.firstPoint & 0xFF;
                    } else if (s.x.length != 1) { 
                       continue;
                    }
                }
                int namePos = 0;
                int nameDir = 1;
                short[] xa = s.x;
                short[] ya = s.y;
                int prevX = ((xa[firstPoint] & MAX_COORD) - x)*scale >> MAX_COORD_BITS;
                int prevY = ((ya[firstPoint] & MAX_COORD) - y)*scale >> MAX_COORD_BITS;
                int len = 0;
                int maxLen = 0;
                int maxFirstPoint = 0;

                if (xa.length == 1) { 
                    // point object
                    g.setColor(OBJECT_COLOR);
                    if (prevX >= 0 && prevX < screenWidth && prevY >= 0 && prevY < screenHeight) { 
                        if (scale < printNamesThreshold) {
                            g.drawLine(prevX, prevY, prevX, prevY);
                        } else if (printNames) { 
                            int si = street.segments.length; 
                            while (--si >= 0 && (street.segments[si] == s || street.segments[si].marker != marker));
                            Image icon = main.objectTypeIcons[street.type];
                            if (icon != null) {  
                                g.drawImage(icon, prevX, prevY, Graphics.HCENTER|Graphics.VCENTER);
                            } else { 
                                if (level == main.nLevels-1) { 
                                    if (si < 0) { 
                                        g.drawString(name, prevX, prevY, Graphics.LEFT|Graphics.BASELINE);  
                                    } else { 
                                        g.drawChar(name.charAt(0), prevX, prevY, Graphics.LEFT|Graphics.BASELINE);  
                                    }
                                } else { 
                                    if (si < 0) { 
                                        g.drawChar(name.charAt(0), prevX, prevY, Graphics.LEFT|Graphics.BASELINE);  
                                    }
                                }
                            }                                    
                        }
                    }
                    continue;
                } else if (street.type >= ObjectType.FIRST_REGION && street.type <= ObjectType.LAST_REGION) { 
                    if (!printNames) {
                        int[] lxa = new int[MAX_PIECES];
                        g.setColor(regionColor(street));
                        /*
                        for (int ly = 0; ly < screenHeight; ly++) { 
                            int n = 0;
                            int x1 = prevX;
                            int y1 = prevY;
                            for (int p = 1; p < xa.length; p++) { 
                                int x0 = x1;
                                int y0 = y1;
                                x1 = ((xa[p] & MAX_COORD) - x)*scale >> MAX_COORD_BITS;
                                y1 = ((ya[p] & MAX_COORD) - y)*scale >> MAX_COORD_BITS;
                                if (y0 != y1 && ((ly > y0 && ly <= y1) || (ly >= y1 && ly < y0))) { 
                                    int lx = x0 + (x1-x0)*(ly-y0)/(y1-y0);
                                    int l = 0, r = n;
                                    while (l < r) { 
                                        int m = (l + r) >>> 1;
                                        if (lxa[m] < lx) { 
                                            l = m + 1;
                                        } else { 
                                            r = m;
                                        }
                                    }
                                    System.arraycopy(lxa, r, lxa, r+1, n-r);
                                    lxa[r] = lx;
                                    n += 1;
                                }
                            }
                            for (int l = 1; l < n; l+=2) { 
                                g.drawLine(lxa[l-1], ly, lxa[l], ly); 
                            }
                        }
                        */
                        for (int sy = 0; sy < screenHeight; sy++) { 
                            int n = 0;
                            int ly = y + (sy << MAX_COORD_BITS) / scale;
                            int x1 = (xa[0] & MAX_COORD);
                            int y1 = (ya[0] & MAX_COORD);
                            for (int p = 1; p < xa.length; p++) { 
                                int x0 = x1;
                                int y0 = y1;
                                x1 = (xa[p] & MAX_COORD);
                                y1 = (ya[p] & MAX_COORD);
                                if (y1 == sy || ((ly - y0) ^ (ly - y1)) < 0) { 
                                    int lx = y1 == sy ? x0 : x0 + (x1-x0)*(ly-y0)/(y1-y0);
                                    int l = 0, r = n;
                                    while (l < r) { 
                                        int m = (l + r) >>> 1;
                                        if (lxa[m] < lx) { 
                                            l = m + 1;
                                        } else { 
                                            r = m;
                                        }
                                    }
                                    if (y1 == sy && ((y1-y0)^((ya[p%(ya.length-1)+1]&MAX_COORD)-y1)) < 0) {
                                        System.arraycopy(lxa, r, lxa, r+2, n-r);
                                        lxa[r] = lxa[r+1] = lx;
                                        n += 2;
                                    }
                                    System.arraycopy(lxa, r, lxa, r+1, n-r);
                                    lxa[r] = lx;
                                    n += 1;
                                }
                            }
                            for (int l = 1; l < n; l+=2) { 
                                g.drawLine((lxa[l] - x)*scale >> MAX_COORD_BITS, sy, 
                                           (lxa[l-1] - x)*scale >> MAX_COORD_BITS, sy); 
                            }
                        }
                        /**/
                    }                                
                    continue;
                }
                
                if (printNames) { 
                    g.setColor(NAME_COLOR);
                } else if (route == null) { 
                    g.setColor(s.streetIndex == selectedStreetIndex ? SELECTION_COLOR : ROAD_COLOR); 
                }
                int cx = -1, cy = -1;
                for (int p = firstPoint+1; p < xa.length; p++) { 
                    int x0 = prevX;
                    int y0 = prevY;
                    int x1 = ((xa[p] & MAX_COORD) - x)*scale >> MAX_COORD_BITS;
                    int y1 = ((ya[p] & MAX_COORD) - y)*scale >> MAX_COORD_BITS;
                    prevX = x1;
                    prevY = y1;                    

                  setColor:
                    if (!printNames && route != null) { 
                        Edge[] re = route.edges;
                        Cross c0 = route.start;
                        for (int m = 0; m < re.length; m++) { 
                            Cross c1 = re[m].target;
                            if (re[m].segment == s) { 
                                for (int l = p; --l >= 0;) {                                         
                                    if (xa[l] == c1.x && ya[l] == c1.y) {  
                                        for (l = p; l < xa.length; l++) { 
                                            if (xa[l] == c0.x && ya[l] == c0.y) { 
                                                g.setColor(ROUTE_COLOR);                                                       
                                                break setColor;
                                            }
                                        }
                                        break;
                                    } else if (xa[l] == c0.x && ya[l] == c0.y) { 
                                        for (l = p; l < xa.length; l++) { 
                                            if (xa[l] == c1.x && ya[l] == c1.y) { 
                                                g.setColor(ROUTE_COLOR);                         
                                                break setColor;
                                            }
                                        }
                                        break;
                                    }
                                }                                        
                            }
                            c0 = c1;
                        }
                        g.setColor(s.streetIndex == selectedStreetIndex ? SELECTION_COLOR : ROAD_COLOR);                         
                    } 

                    if (scale < printNamesThreshold) { 
                        g.drawLine(x0, y0, x1, y1); 
                        continue;
                    }
                    if (x0 < 0) { 
                        len = 0;
                        firstPoint = p-1;
                        if (x1 < 0) { 
                            continue;
                        }
                        y0 += (y1-y0)*(-x0)/(x1-x0);
                        x0 = 0;
                    } else if (x0 > screenWidth) { 
                        len = 0;
                        firstPoint = p-1;
                        if (x1 > screenWidth) {
                            continue;
                        }
                        y0 += (y1-y0)*(screenWidth-x0)/(x1-x0);
                        x0 = screenWidth;
                    }
                    if (x1 < 0) { 
                        y1 += (y0-y1)*(-x1)/(x0-x1);
                        x1 = 0;
                    } else if (x1 > screenWidth) { 
                        y1 += (y0-y1)*(screenWidth-x1)/(x0-x1);
                        x1 = screenWidth;
                    }
                    if (y0 < 0) { 
                        len = 0;
                        firstPoint = p-1;
                        if (y1 < 0) { 
                            continue;
                        }
                        x0 += (x1-x0)*(-y0)/(y1-y0);
                        y0 = 0;
                    } else if (y0 > screenHeight) { 
                        len = 0;
                        firstPoint = p-1;
                        if (y1 > screenHeight) {
                            continue;
                        }
                        x0 += (x1-x0)*(screenHeight-y0)/(y1-y0);
                        y0 = screenHeight;
                    }
                    if (y1 < 0) { 
                        x1 += (x0-x1)*(-y1)/(y0-y1);
                        y1 = 0;
                    } else if (y1 > screenHeight) { 
                        x1 += (x0-x1)*(screenHeight-y1)/(y0-y1);
                        y1 = screenHeight;
                    }             
                    if (!printNames) { 
                        g.drawLine(x0, y0, x1, y1); 
                    }
                    int width = level - s.level + 1;
                    int dx = 0;
                    int dy = 0;                    
                    if (Math.abs(x1-x0) >= Math.abs(y1-y0)) { 
                        // horizontal line
                        if (!printNames) { 
                            do { 
                                dy += 1;
                                g.drawLine(x0, y0-dy, x1, y1-dy); 
                                g.drawLine(x0, y0+dy, x1, y1+dy); 
                            } while (--width != 0); 
                        
                            len += Math.abs(x1-x0);
                            if (len-indent > maxLen) {
                                maxLen = len-indent;
                                maxFirstPoint = firstPoint;
                            }
                        } else if (x0 != x1) {                             
                            if (Locale.HORIZONTAL_STREET_NAMES) { 
                                if (namePos == 0) { 
                                    g.drawString(name, 
                                                 (x0 + x1) >> 1,
                                                 (y0 + y1) >> 1, 
                                                 Graphics.LEFT|Graphics.BASELINE); 
                                    namePos = nameLen;
                                }
                            } else { 
                                dx = x0 <= x1 ? 1 : -1; 
                                if (namePos == 0) { 
                                    cx = x0 + dx*indent;
                                    if (dx < 0 && ((cx - x0) ^ (cx - x1)) < 0) { 
                                        int suffixLen = nameLen;
                                        while (font.substringWidth(name, nameLen-suffixLen, suffixLen) > s.maxLength) { 
                                            suffixLen -= 1;
                                        }
                                        namePos = suffixLen-1;
                                        nameDir = -1;
                                    }
                                } else if (cx < 0 || ((cx - x0) ^ (cx - x1)) >= 0) { 
                                    cx = x0 + dx;
                                    // cx = x0 + (dx*fontSpaceWidth >> 1);
                                }
                                int nChars = 0;
                                while (namePos < nameLen && namePos >= 0 && ((cx - x0) ^ (cx - x1)) < 0) { 
                                    cy = y0 + (y1-y0)*(cx-x0)/(x1-x0);
                                    char ch = name.charAt(namePos);
                                    if (!isOccupied(cx, cy)) { 
                                        g.drawChar(ch, 
                                                   cx,
                                                   cy + (fontHeight >> 1), 
                                                   Graphics.HCENTER|Graphics.BOTTOM); 
                                        namePos += nameDir;
                                        nChars += 1;
                                    } else { 
                                        int np = namePos;
                                        int px = cx;
                                        while (nChars != 0) { 
                                            nChars -= 1;
                                            np -= nameDir;
                                            char pc = name.charAt(np);
                                            px -= dx*font.charWidth(pc);
                                            setOccupied(px, y0 + (y1-y0)*(px-x0)/(x1-x0));
                                        }
                                    }
                                    cx += dx*font.charWidth(ch);
                                } 
                                int np = namePos;
                                int px = cx;
                                while (nChars != 0) { 
                                    nChars -= 1;
                                    np -= nameDir;
                                    char pc = name.charAt(np);
                                    px -= dx*font.charWidth(pc);
                                    setOccupied(px, y0 + (y1-y0)*(px-x0)/(x1-x0));
                                }
                            }
                        }
                    } else { 
                        // vertical line
                        if (!printNames) { 
                            do { 
                                dx += 1;
                                g.drawLine(x0-dx, y0, x1-dx, y1); 
                                g.drawLine(x0+dx, y0, x1+dx, y1); 
                            } while (--width != 0); 

                            len += Math.abs(y1-y0);
                            if (len > maxLen) {
                                maxLen = len;
                                maxFirstPoint = firstPoint;
                            }
                        } else if (y0 != y1) { 
                            if (Locale.HORIZONTAL_STREET_NAMES) { 
                                if (namePos == 0) { 
                                    g.drawString(name, 
                                                 (x0 + x1) >> 1,
                                                 (y0 + y1) >> 1, 
                                                 Graphics.LEFT|Graphics.BASELINE); 
                                    namePos = nameLen;
                                }
                            } else { 
                                dy = y0 <= y1 ? 1 : -1; 
                                if (namePos == 0) { 
                                    cy = y0 + indent*dy;
                                    if ((x0 > x1 || (x0 == x1 && dy < 0)) && ((cy - y0) ^ (cy - y1)) < 0) { 
                                        int suffixLen = (s.maxLength + fontHeight - 1) / fontHeight;
                                        namePos = (suffixLen < nameLen) ? suffixLen-1 : nameLen-1;
                                        nameDir = -1;
                                    }
                                } else if (cy < 0 || ((cy - y0) ^ (cy - y1)) >= 0) { 
                                    cy = y0 + dy;
                                }
                                int nChars = 0;
                                while (namePos < nameLen && namePos >= 0 && ((cy - y0) ^ (cy - y1)) < 0) { 
                                    cx = x0 + (x1-x0)*(cy-y0)/(y1-y0);
                                    if (!isOccupied(cx, cy)) { 
                                        g.drawChar(name.charAt(namePos), 
                                                   cx,
                                                   cy + (fontHeight >> 1), 
                                                   Graphics.HCENTER|Graphics.BOTTOM); 
                                        namePos += nameDir;
                                        nChars += 1;
                                    } else { 
                                        int py = cy;
                                        while (nChars != 0) { 
                                            nChars -= 1;
                                            py -= dy*fontHeight;
                                            setOccupied(x0 + (x1-x0)*(py-y0)/(y1-y0), py);
                                        }
                                    }
                                    cy += dy*fontHeight;
                                }
                                int py = cy;
                                while (nChars != 0) { 
                                    nChars -= 1;
                                    py -= dy*fontHeight;
                                    setOccupied(x0 + (x1-x0)*(py-y0)/(y1-y0), py);
                                }
                            }
                        }
                    }
                }
                s.maxLength = (short)maxLen;
                s.firstPoint = (byte)maxFirstPoint;
            }
        }
    }
                    

    public void pointerPressed(int x, int y) { 
        cx = x;
        cy = y;
        repaint();
    }

    public void pointerReleased(int x, int y) { 
        pointerDragged(x, y);
    }

    public void pointerDragged(int x, int y) { 
        this.x -= ((x - cx) << MAX_COORD_BITS)/scale;
        this.y -= ((y - cy) << MAX_COORD_BITS)/scale;
        cx = x;
        cy = y;        
        checkCoordinates();
        repaint();
    }


    public void keyPressed(int keyCode) {
        if (hasPointerEvents()) { 
            keyAction(keyCode);
        } else { 
            if (repeatTask != null) { 
                repeatTask.cancel();
            }
            repeatTask = new RepeatTask(keyCode);
            repeatTimer.schedule(repeatTask, REPEAT_DELAY, REPEAT_PERIOD);
        }
    }

    public void keyReleased(int keyCode) {
        if (!hasPointerEvents()) { 
            keyAction(keyCode);
            if (repeatTask != null) { 
                repeatTask.cancel();
                repeatTask = null;
            }
        }
    }

    private void keyAction(int keyCode) {
        int step = size >> 2;

        switch (keyCode) {
        case KEY_NUM0:
            reset();
            break;
        case KEY_NUM1:
            x -= step;
            y -= step;
            break;
        case KEY_NUM2:
            y -= step;
            break;
        case KEY_NUM3:
            x += step;
            y -= step;
            break;
        case KEY_NUM4:
            x -= step;
            break;
        case KEY_NUM5:
            new SearchBox(this);
            return;
        case KEY_NUM6:
            x += step;
            break;
        case KEY_NUM7:
            x -= step;
            y += step;
            break;
        case KEY_NUM8:
            y += step;
            break;
        case KEY_NUM9:
            x += step;
            y += step;
            break;
        case KEY_STAR:
            increase();
            return;
        case KEY_POUND:
            decrease();
            return;
        default:
            switch (getGameAction(keyCode)) { 
            case GAME_A:
                increase();
                return;                
            case GAME_B:
                decrease();
                return;                
            case LEFT:
                cx -= CURSOR_STEP;
                break;
            case RIGHT:
                cx += CURSOR_STEP;
                break;
            case UP:
                cy -= CURSOR_STEP;
                break;
            case DOWN:
                cy += CURSOR_STEP;
                break;
            case FIRE:
                fire();
                // no break
            default:
                return;
            }
        }
        checkCoordinates();
        repaint();
    }

    Cross searchNearestCross(int x, int y) { 
        Cross[] crosses = main.crosses;
        Cross nearestCross = null;
        long minDistance = Long.MAX_VALUE;
        for (int i = 0; i < crosses.length; i++) { 
            Cross c = crosses[i];
            long d = (long)(x - (c.x & MAX_COORD))*(x - (c.x & MAX_COORD)) 
                + (long)(y - (c.y & MAX_COORD))*(y - (c.y & MAX_COORD));
            if (d < minDistance) { 
                minDistance = d;
                nearestCross = c;
            }
        }
        return nearestCross;
    }

    Segment findObjectInfo(int x3, int y3) { 
        int i = x3 >>> (MAX_COORD_BITS - MobileMap.CELL_BITS);
        int j = y3 >>> (MAX_COORD_BITS - MobileMap.CELL_BITS);
        Segment[] segments = main.cells[i][j];
        Segment nearestSegment = null;
        long minDistance = Long.MAX_VALUE;

        for (int k = 0; k < segments.length; k++) { 
            Segment s = segments[k];
            int xc, yc;
            int len = s.x.length;
            if (len == 1) { 
                xc = s.x[0] & MAX_COORD;
                yc = s.y[0] & MAX_COORD;
                long d = (long)(x3 - xc)*(x3 - xc) + (long)(y3 - yc)*(y3 - yc);
                if (d < minDistance) { 
                    minDistance = d;
                    nearestSegment = s;
                }
            } else if (len > 1) { 
                int x2 = s.x[0] & MAX_COORD;
                int y2 = s.y[0] & MAX_COORD;
                for (int l = 1; l < len; l++) { 
                    int x1 = x2;
                    int y1 = y2;
                    x2 = s.x[l] & MAX_COORD;
                    y2 = s.y[l] & MAX_COORD;
                    if (x2 == x1 && y2 == y1) { 
                        System.out.println("Zero coordinates for street " + main.streets[s.streetIndex].name);
                        continue;
                    }
                    yc = (int)(((x2-x1)*(y2-y1)*(long)(x3-x1) + (y2-y1)*(y2-y1)*(long)y3 + (x2-x1)*(x2-x1)*(long)y1)/
                                   ((x2-x1)*(x2-x1) + (y2-y1)*(y2-y1)));
                    xc = (x1 != x2)
                        ? (y3-yc)*(y2-y1)/(x2-x1) + x3
                        : (yc-y1)*(x2-x1)/(y2-y1) + x1;

                    if ((xc-x1)*(xc-x2) <= 0 && (yc-y1)*(yc-y2) <= 0) {
                        long d = (x3 - xc)*(x3 - xc) + (y3 - yc)*(y3 - yc);
                        if (d < minDistance) { 
                            minDistance = d;
                            nearestSegment = s;
                        }
                    }
                }
            }
        }
        return nearestSegment;
    }

    void fire() { 
        int fx = x + (cx << MAX_COORD_BITS)/scale;
        int fy = y + (cy << MAX_COORD_BITS)/scale;
        switch (fireMode) { 
        case FIRE_INFO:
        {
            Segment s = findObjectInfo(fx, fy);
            if (s != null) { 
                Street street = main.streets[s.streetIndex];
                showInfo(street, fx, fy);
            } else { 
                showError(Locale.NO_OBJECT);
            }
            break;
        }
        case FIRE_START_POINT:
            startPoint = searchNearestCross(fx, fy);
            route = null;
            if (startPoint != null) { 
                fireMode = FIRE_INFO;
                repaint();
            } else { 
                showError(Locale.NO_OBJECT);
            }
            break;
        case FIRE_FINISH_POINT:
            finishPoint = searchNearestCross(fx, fy);
            route = null;
            if (finishPoint != null) { 
                fireMode = FIRE_INFO;
                repaint();
            } else { 
                showError(Locale.NO_OBJECT);
            }
            break;
        }
    }
    public void commandAction(Command c, Displayable d)  
    {
        if (c == MobileMap.QUIT_CMD) {
            main.quit();
        } else if (c == MobileMap.ZOOM_IN_CMD) { 
            increase();
        } else if (c == MobileMap.ZOOM_OUT_CMD) { 
            decrease();
        } else if (c == MobileMap.RESET_CMD) { 
            reset();
            repaint();            
        } else if (c == MobileMap.SEARCH_CMD) { 
            new SearchBox(this);
        } else if (c == MobileMap.INDEX_CMD) { 
            new IndexList(this);
        } else if (c == MobileMap.START_POINT_CMD) { 
            fireMode = FIRE_START_POINT;
            repaint();
        } else if (c == MobileMap.FINISH_POINT_CMD) { 
            fireMode = FIRE_FINISH_POINT;
            repaint();
        } else if (c == MobileMap.FASTEST_ROUTE_CMD) { 
            if (startPoint == null) { 
                showError(Locale.START_POINT_IS_NOT_SET);
            } else if (finishPoint == null) { 
                showError(Locale.FINISH_POINT_IS_NOT_SET);
            } else { 
                route = main.findFastestRoute(startPoint, finishPoint);
                if (route == null) { 
                    removeCommand(MobileMap.EXPLAIN_ROUTE_CMD);
                    showError(Locale.NO_ROUTE);
                } else { 
                    addCommand(MobileMap.EXPLAIN_ROUTE_CMD);
                    new RouteForm(this, Locale.FASTEST_ROUTE);
                }
            }                
        } else if (c == MobileMap.SHORTEST_ROUTE_CMD) { 
            if (startPoint == null) { 
                showError(Locale.START_POINT_IS_NOT_SET);
            } else if (finishPoint == null) { 
                showError(Locale.FINISH_POINT_IS_NOT_SET);
            } else { 
                route = main.findShortestRoute(startPoint, finishPoint);
                if (route == null) { 
                    removeCommand(MobileMap.EXPLAIN_ROUTE_CMD);
                    showError(Locale.NO_ROUTE);
                } else { 
                    addCommand(MobileMap.EXPLAIN_ROUTE_CMD);
                    new RouteForm(this, Locale.SHORTEST_ROUTE);
                }
            }                
        } else if (c == MobileMap.EXPLAIN_ROUTE_CMD) {
            if (route != null) { 
                new RouteList(this);
            }
        } else if (c == MobileMap.HELP_CMD) { 
            showHelp();
        }
    }

    void showHelp() { 
	new HelpForm(this);
    }

    void showInfo(Street street, int fx, int fy) { 
        new InfoForm(this, street, fx, fy);
    }

    void showError(String msg) { 
	Alert alert = new Alert(Locale.ERROR, msg, null, AlertType.ERROR);
	Display.getDisplay(main).setCurrent(alert, this);
    }
        
    void checkCoordinates() { 
        int shift = size >> 1;
        int step = shift*scale >> MAX_COORD_BITS;

        if (cx < 0) { 
            cx += step;
            x -= shift;
        } else if (cx >= screenWidth) { 
            cx -= step;
            x += shift;
        }
        if (cy < 0) { 
            cy += step;
            y -= shift;
        } else if (cy >= screenHeight) { 
            cy -= step;
            y += shift;
        }

        if (x < 0) { 
            x = 0;
        } else if (x + size > MAX_COORD) { 
            x = MAX_COORD - size;
        }
        if (y < 0) { 
            y = 0;
        } else if (y + size > MAX_COORD) { 
            y = MAX_COORD - size;
        }
    }

    void reset() { 
        scale = initScale;
        x = y = 0;
        cx = screenWidth/2;
        cy = screenHeight/2;
        size = MAX_COORD;
        level = 0;
        selectedStreetIndex = -1;
        route = null;
        fireMode = FIRE_INFO;
    }        

    void selectStreet(int index) { 
        selectedStreetIndex = index;
        if (index >= 0) { 
            Street street = main.streets[index];
            Segment[] segments = street.segments;
            int minX = Integer.MAX_VALUE;
            int minY = Integer.MAX_VALUE;
            int maxX = Integer.MIN_VALUE;
            int maxY = Integer.MIN_VALUE;
            int streetLevel = 0;
            for (int i = 0; i < segments.length; i++) { 
                Segment s = segments[i];
                streetLevel = s.level;
                for (int j = 0; j < s.x.length; j++) { 
                    int x = s.x[j] & 0xFFFF;
                    int y = s.y[j] & 0xFFFF;
 
                    if (x < minX) { 
                        minX = x;
                    } 
                    if (x > maxX) { 
                        maxX = x;
                    } 
                    if (y < minY) { 
                        minY = y;
                    } 
                    if (y > maxY) { 
                        maxY = y;
                    } 
                }
            }
            int width = maxX - minX;
            int height = maxY - minY;
            if (width == 0 && height == 0) { 
                x = minX - (size >> 1);
                y = minY - (size >> 1);
                checkCoordinates();                    
                cx = (minX - x)*scale >> MAX_COORD_BITS;
                cy = (minY - y)*scale >> MAX_COORD_BITS;
            } else { 
                int wrapSize = (width > height ? width : height) << 1;
                for (size = MAX_COORD, scale = initScale; size > wrapSize; size >>= 1, scale <<= 1);
                for (setLevel(); level < streetLevel;  size >>= 1, scale <<= 1, setLevel());
                x = minX + (width >> 1) - (size >> 1);
                y = minY + (height >> 1) - (size >> 1);                  
                checkCoordinates();                    
            }
        }
        repaint();
    }
       

    void setLevel() { 
        level = 0;
        for (int i = 1; i < minLevelScale.length && minLevelScale[i] < scale; i++) { 
            level += 1;
        }
    }

    void increase() { 
        if ((scale >>> (MAX_COORD_BITS-2)) == 0) {             
            int fx = x + (cx << MAX_COORD_BITS)/scale;
            int fy = y + (cy << MAX_COORD_BITS)/scale;
            scale <<= 1;
            setLevel();
            size >>= 1;
            x = fx - (size >> 1);
            y = fy - (size >> 1);
            checkCoordinates();
            cx = (fx - x)*scale >> MAX_COORD_BITS;
            cy = (fy - y)*scale >> MAX_COORD_BITS;            
            repaint();
        }
    }

    void decrease() { 
        int fx = x + (cx << MAX_COORD_BITS)/scale;
        int fy = y + (cy << MAX_COORD_BITS)/scale;
        scale >>= 1;
        if (scale <= initScale) { 
            reset();
        } else { 
            setLevel();
            x -= (size >> 1);
            y -= (size >> 1);
            size <<= 1;
        }
        checkCoordinates();
        cx = (fx - x)*scale >> MAX_COORD_BITS;
        cy = (fy - y)*scale >> MAX_COORD_BITS;            
        repaint();
    }
}
            
