import javax.microedition.lcdui.*;
import javax.microedition.midlet.*;
import java.util.*;
import java.io.*;

public class MobileMap extends MIDlet
{ 
    static final Command BACK_CMD = new Command(Locale.BACK, Command.BACK, 1);
    static final Command QUIT_CMD = new Command(Locale.QUIT, Command.BACK, 2);
    static final Command SEARCH_CMD = new Command(Locale.SEARCH, Command.SCREEN, 1);
    static final Command INDEX_CMD = new Command(Locale.INDEX, Command.SCREEN, 2);
    static final Command ZOOM_IN_CMD = new Command(Locale.ZOOM_IN, Command.SCREEN, 3);
    static final Command ZOOM_OUT_CMD = new Command(Locale.ZOOM_OUT, Command.SCREEN, 3);
    static final Command EXPLAIN_ROUTE_CMD = new Command(Locale.EXPLAIN_ROUTE, Command.SCREEN, 4);
    static final Command START_POINT_CMD = new Command(Locale.START_POINT, Command.SCREEN, 5);
    static final Command FINISH_POINT_CMD = new Command(Locale.FINISH_POINT, Command.SCREEN, 5);
    static final Command FASTEST_ROUTE_CMD = new Command(Locale.FASTEST_ROUTE, Command.SCREEN, 6);
    static final Command SHORTEST_ROUTE_CMD = new Command(Locale.SHORTEST_ROUTE, Command.SCREEN, 7);
    static final Command HELP_CMD = new Command(Locale.HELP, Command.SCREEN, 8);
    static final Command RESET_CMD = new Command(Locale.RESET, Command.SCREEN, 9);

    static final String MAP_FILE = "/city.map";
    static final int CELL_BITS = 3;
    static final int N_CELLS = 1 << CELL_BITS;

    final Image[] objectTypeIcons;
    final int[]   nObjects;

    int nLevels;
    int nLevelStreets[];

    Street[]      streets;
    Segment[][][] cells;
    Cross[]       crosses;

    int findStreet(String prefix) { 
        int l = 0, n = streets.length, r = n; 
        while (l < r) { 
            int m = (l + r) >>> 1;
            if (prefix.compareTo(streets[m].name) > 0) { 
                l = m + 1;
            } else { 
                r = m;
            }
        }
        if (r == n || !streets[r].name.startsWith(prefix)) { 
            for (int i = 0; i < n; i++) { 
                if (streets[i].name.indexOf(prefix) >= 0) { 
                    return i;
                }
            }
        }
        return r;
    }

    public MobileMap() throws Exception { 
        objectTypeIcons = new Image[ObjectType.name.length];
        nObjects = new int[ObjectType.name.length];
        for (int i = 0; i < ObjectType.name.length; i++) { 
            try { 
                objectTypeIcons[i] = Image.createImage("/" + ObjectType.name[i] + ".png");
            } catch (IOException x) {}
        }
        DataInputStream in = new DataInputStream(getClass().getResourceAsStream(MAP_FILE));
        int nStreets = in.readShort() & 0xFFFF;
        streets = new Street[nStreets];
        for (int i = 0; i < nStreets; i++) { 
            Street street = new Street();
            int nSegments = in.readByte() & 0xFF;
            street.segments = new Segment[nSegments];
            street.type = in.readByte();
            nObjects[street.type] += 1;
            int len = in.readByte() & 0xFF;
            byte[] name = new byte[len];
            for (int j = 0; j < len; j++) { 
                name[j] = in.readByte();
            }
            street.name = Locale.decodeString(name);
            streets[i] = street;
        }
        cells = new Segment[N_CELLS][N_CELLS][];
         for (int i = 0; i < N_CELLS; i++) { 
            for (int j = 0; j < N_CELLS; j++) { 
                cells[i][j] = new Segment[in.readShort() & 0xFFFF];
            }
        }
        nLevels = in.readByte();
        nLevelStreets = new int[nLevels];
        int[][] cellPos = new int[N_CELLS][N_CELLS];
        short[] segmentPos = new short[nStreets];
        for (int i = 0; i < nLevels; i++) { 
            int nSegments = in.readShort() & 0xFFFF; 
            nLevelStreets[i] = nSegments;
            for (int j = 0; j < nSegments; j++) {            
                Segment s = new Segment();
                s.streetIndex = in.readShort();
                /*
                if (segmentPos[s.streetIndex] ==  streets[s.streetIndex].segments.length) { 
                    System.out.println("nSegments = " + nSegments + ", j=" + j + ", segmentPos[s.streetIndex]=" + segmentPos[s.streetIndex]);
                }
                */
                streets[s.streetIndex].segments[segmentPos[s.streetIndex]++] = s;
                s.level = (byte)i;
                int nPoints = in.readByte() & 0xFF;
                s.x = new short[nPoints];
                s.y = new short[nPoints];
                int minX = Integer.MAX_VALUE;
                int minY = Integer.MAX_VALUE;
                int maxX = Integer.MIN_VALUE;
                int maxY = Integer.MIN_VALUE;
                for (int k = 0; k < nPoints; k++) {                 
                    int x = in.readShort() & MapCanvas.MAX_COORD;
                    int y = in.readShort() & MapCanvas.MAX_COORD;
                    if (x < minX) { 
                        minX = x;
                    }
                    if (x > maxX) { 
                        maxX = x;
                    }
                    if (y < minY) { 
                        minY = y;
                    }
                    if (y > maxY) { 
                        maxY = y;
                    }
                    s.x[k] = (short)x;
                    s.y[k] = (short)y;
                }
                minX >>>= MapCanvas.MAX_COORD_BITS-CELL_BITS;
                maxX >>>= MapCanvas.MAX_COORD_BITS-CELL_BITS;
                minY >>>= MapCanvas.MAX_COORD_BITS-CELL_BITS;
                maxY >>>= MapCanvas.MAX_COORD_BITS-CELL_BITS;
                for (int cx = minX; cx <= maxX; cx++) { 
                    for (int cy = minY; cy <= maxY; cy++) { 
                        cells[cx][cy][cellPos[cx][cy]++] = s;
                    }
                }
            }
        }
        int nCrosses = in.readShort() & 0xFFFF;
        Cross[] crosses = new Cross[nCrosses];
        this.crosses = crosses;
        for (int i = 0; i < nCrosses; i++) { 
            crosses[i] = new Cross();
        }
        for (int i = 0; i < nCrosses; i++) { 
            Cross c = crosses[i];
            c.x = in.readShort();
            c.y = in.readShort();
            int nNeighbors = in.readByte() & 0xFF;
            c.neighbors = new Edge[nNeighbors];
            for (int j = 0; j < nNeighbors; j++) { 
                Edge e = new Edge();
                c.neighbors[j] = e;
                e.target = crosses[in.readShort() & 0xFFFF];
                Street street = streets[in.readShort() & 0xFFFF];
                e.segment =  street.segments[in.readByte() & 0xFF];
                e.distance = in.readShort();
                e.time = in.readShort();
            }
        }
        in.close();
    }

    Route findShortestRoute(Cross from, Cross till) { 
        Cross[] crosses = this.crosses;
        for (int i = 0; i < crosses.length; i++) { 
            crosses[i].weight = Integer.MAX_VALUE;
        }
        from.weight = 0;
        till.weight = ~0;
        Queue queue = new Queue();
        queue.enqueue(from);

        System.out.println("Start point at:");
        for (int i = 0; i < from.neighbors.length; i++) { 
            System.out.print(" " + streets[from.neighbors[i].segment.streetIndex].name);
        }
        System.out.println("Finish point at:");
        for (int i = 0; i < till.neighbors.length; i++) { 
            System.out.print(" " + streets[till.neighbors[i].segment.streetIndex].name);
        }
        System.out.println();

        while (queue.size() != 0) { 
            Cross c = (Cross)queue.dequeue();
            c.weight &= Integer.MAX_VALUE;
            Edge[] neighbors = c.neighbors;
            for (int i = 0; i < neighbors.length; i++) { 
                Cross nc = neighbors[i].target;
                int newDistance = c.weight + (neighbors[i].distance & 0xFFFF);
                if (newDistance < (till.weight & Integer.MAX_VALUE) 
                    && newDistance < (nc.weight & Integer.MAX_VALUE)) 
                { 
                    if (nc.weight >= 0) { 
                        queue.enqueue(nc);
                    }
                    if (c == nc) throw new Error();
                    nc.weight = Integer.MIN_VALUE|newDistance;
                    nc.back = c;
                }
            }
        }
        if (till.weight == ~0) { 
            // path not found
            return null;
        }
        till.weight &= Integer.MAX_VALUE;

        Vector v = new Vector();
        Cross dst = till;
        int time = 0;
        //System.out.println("till.x = " + (till.x & 0xFFFF) + ", till.y=" + (till.y & 0xFFFF) + ", till.weigth=" + till.weight);

        do { 
            Cross src = dst.back;
            Edge[] se = src.neighbors;
            int i;
            for (i = 0; dst != se[i].target; i++);
            Edge e = se[i];
            //System.out.println("Edge name=" + streets[e.segment.streetIndex].name + ", edge distance " + (e.distance & 0xFFFF) + ", src.weight=" + src.weight + "src.x = " + (src.x & 0xFFFF) + ", src.y=" + (src.y & 0xFFFF) + ", src.neighbors.length=" + src.neighbors.length);            
            //for (int j = 0; j < src.neighbors.length; j++) System.out.println("Neighbor x=" + (src.neighbors[j].target.x & 0xFFFF) + ", y=" + (src.neighbors[j].target.y & 0xFFFF));
            
            v.addElement(e);
            time += e.time & 0xFFFF;
            dst = src;
        } while (dst != from);

        int n = v.size();
        Edge[] de = new Edge[n];
        for (int i = 0; i < n; i++) { 
            de[i] = (Edge)v.elementAt(n-i-1);
        }
        return new Route(from, till, de, till.weight, time);
    }


    
    Route findFastestRoute(Cross from, Cross till) { 
        Cross[] crosses = this.crosses;
        for (int i = 0; i < crosses.length; i++) { 
            crosses[i].weight = Integer.MAX_VALUE;
        }
        from.weight = 0;
        till.weight = ~0;
        Queue queue = new Queue();
        queue.enqueue(from);
        while (queue.size() != 0) { 
            Cross c = (Cross)queue.dequeue();
            c.weight &= Integer.MAX_VALUE;
            Edge[] neighbors = c.neighbors;
            for (int i = 0; i < neighbors.length; i++) { 
                Cross nc = neighbors[i].target;
                int newTime = c.weight + (neighbors[i].time & 0xFFFF);
                if (newTime < (nc.weight & Integer.MAX_VALUE)
                    && newTime < (nc.weight & Integer.MAX_VALUE)) 
                { 
                    if (nc.weight >= 0) { 
                        queue.enqueue(nc);
                    }
                    nc.weight = Integer.MIN_VALUE|newTime;
                    nc.back = c;
                }
            }
        }
        if (till.weight == ~0) { 
            // path not found
            return null;
        }
        till.weight &= Integer.MAX_VALUE;
        //System.out.println("till.x = " + (till.x & 0xFFFF) + ", till.y=" + (till.y & 0xFFFF) + ", till.weigth=" + till.weight);

        Vector v = new Vector();
        Cross dst = till;
        int distance = 0;
        do { 
            Cross src = dst.back;
            Edge[] se = src.neighbors;
            int i;
            for (i = 0; dst != se[i].target; i++);
            Edge e = se[i];
            //System.out.println("edge time " + (e.time & 0xFFFF) + ", src.weight=" + src.weight + "src.x = " + (src.x & 0xFFFF) + ", src.y=" + (src.y & 0xFFFF));
            v.addElement(e);
            distance += e.distance & 0xFFFF;
            dst = src;
        } while (dst != from);

        int n = v.size();
        Edge[] de = new Edge[n];
        for (int i = 0; i < n; i++) { 
            de[i] = (Edge)v.elementAt(n-i-1);
        }
        return new Route(from, till, de, distance, till.weight & Integer.MAX_VALUE);
    }
    

    protected void destroyApp(boolean unconditional) {
    }    

    protected  void pauseApp() {
    }

    protected void startApp() 
    {
        new MapCanvas(this);
    }

    public void quit() { 
        destroyApp(true);
        notifyDestroyed();
    }

}    
 